from typing import List

import pytest
from aiostream import stream

from resotocore.cli import strip_quotes, js_value_at
from resotocore.cli.cli import multi_command_parser, CLIService
from resotocore.cli.command import (
    ExecuteSearchCommand,
    ChunkCommand,
    FlattenCommand,
    UniqCommand,
    EchoCommand,
    AggregateToCountCommand,
    PredecessorsPart,
    DumpCommand,
)
from resotocore.cli.model import (
    ParsedCommands,
    ParsedCommand,
    CLIContext,
    CLI,
    InfraAppAlias,
    CLICommand,
    ExecutableCommand,
)
from resotocore.error import CLIParseError
from resotocore.model.graph_access import EdgeTypes
from resotocore.util import utc


def test_strip() -> None:
    assert strip_quotes("'test'") == "test"
    assert strip_quotes('"test"') == "test"


def test_command_line_parser() -> None:
    def check(cmd_lind: str, expected: List[List[str]]) -> None:
        parsed: List[ParsedCommands] = multi_command_parser.parse(cmd_lind)

        def cmd_to_str(cmd: ParsedCommand) -> str:
            arg = f" {cmd.args}" if cmd.args else ""
            return f"{cmd.cmd}{arg}"

        assert [[cmd_to_str(cmd) for cmd in line.commands] for line in parsed] == expected

    check("test", [["test"]])
    check("test | bla |  bar", [["test", "bla", "bar"]])
    check('search is(foo) and bla.test=="foo"', [['search is(foo) and bla.test=="foo"']])
    check('a 1 | b "s" | c 1.23 | d', [["a 1", 'b "s"', "c 1.23", "d"]])
    check('jq ". | {a:.foo, b: .bla}" ', [['jq ". | {a:.foo, b: .bla}"']])
    check("a|b|c;d|e|f;g|e|h", [["a", "b", "c"], ["d", "e", "f"], ["g", "e", "h"]])
    check("add_job 'what \" test | foo | bla'", [["add_job 'what \" test | foo | bla'"]])
    check('add_job what \\" test \\| foo \\| bla', [['add_job what " test | foo | bla']])


@pytest.mark.asyncio
async def test_multi_command(cli: CLI) -> None:
    nums = ",".join([f'{{ "num": {a}}}' for a in range(0, 100)])
    source = "echo [" + nums + "," + nums + "]"
    command1 = f"{source} | chunk 7 | dump"
    command2 = f"{source} | chunk | flatten | uniq | dump"
    command3 = f"{source} | chunk 10 | dump"
    commands = ";".join([command1, command2, command3])
    result = await cli.evaluate_cli_command(commands)
    assert len(result) == 3
    line1, line2, line3 = result
    assert len(line1.commands) == 3
    l1p1, l1p2, _ = line1.commands
    assert isinstance(l1p1, EchoCommand)
    assert isinstance(l1p2, ChunkCommand)
    assert len(line2.commands) == 5
    l2p1, l2p2, l2p3, l2p4, _ = line2.commands
    assert isinstance(l2p1, EchoCommand)
    assert isinstance(l2p2, ChunkCommand)
    assert isinstance(l2p3, FlattenCommand)
    assert isinstance(l2p4, UniqCommand)
    assert len(line3.commands) == 3
    l3p1, l3p2, _ = line3.commands
    assert isinstance(l3p1, EchoCommand)
    assert isinstance(l3p2, ChunkCommand)


@pytest.mark.asyncio
async def test_query_database(cli: CLIService) -> None:
    query = 'search is("foo") and some_string=="hello" --> f>12 and f<100 and g[*]==2'
    commands = "|".join([query, "count f", "dump"])
    result = await cli.evaluate_cli_command(commands)
    assert len(result) == 1
    line1 = result[0]
    assert len(line1.commands) == 3
    p1, p2, p3 = line1.commands
    assert isinstance(p1, ExecuteSearchCommand)
    assert isinstance(p2, AggregateToCountCommand)
    assert isinstance(p3, DumpCommand)

    with pytest.raises(Exception):
        await cli.evaluate_cli_command("search a>>>>")  # command is un-parsable

    with pytest.raises(CLIParseError):
        cli.cli_env = {}  # delete the env
        await cli.evaluate_cli_command("query id==3")  # no graph specified


@pytest.mark.asyncio
async def test_unknown_command(cli: CLI) -> None:
    with pytest.raises(CLIParseError) as ex:
        await cli.evaluate_cli_command("echo foo | uniq |  some_not_existing_command")
    assert str(ex.value) == "Command >some_not_existing_command< is not known. Typo?"


@pytest.mark.asyncio
async def test_order_of_commands(cli: CLI) -> None:
    with pytest.raises(CLIParseError) as ex:
        await cli.evaluate_cli_command("uniq")
    assert str(ex.value) == "Command >uniq< can not be used in this position: no source data given"

    with pytest.raises(CLIParseError) as ex:
        await cli.evaluate_cli_command("echo foo | uniq | search bla==23")
    assert str(ex.value) == "Command >search< can not be used in this position: must be the first command"


@pytest.mark.asyncio
async def test_help(cli: CLI) -> None:
    result = await cli.execute_cli_command("help", stream.list)
    assert len(result[0]) == 1

    # help for command
    result = await cli.execute_cli_command("help count", stream.list)
    assert len(result[0]) == 1

    # help for alias
    result = await cli.execute_cli_command("help kind", stream.list)
    assert len(result[0]) == 1

    # help for alias template
    result = await cli.execute_cli_command("help discord", stream.list)
    assert len(result[0]) == 1

    # help for infra app alias
    cli.register_infra_app_alias(InfraAppAlias("testcommand", "this is a test alias", "this is a readme", []))
    result = await cli.execute_cli_command("help testcommand", stream.list)
    assert len(result[0]) == 1


@pytest.mark.asyncio
async def test_parse_env_vars(cli: CLI) -> None:
    result = await cli.execute_cli_command('test=foo bla="bar"   d=true env', stream.list)
    # the env is allowed to have more items. Check only for this subset.
    assert {"test": "foo", "bla": "bar", "d": True}.items() <= result[0][0].items()


def test_parse_predecessor_successor_ancestor_descendant_args() -> None:
    plain = CLIContext()
    w_delete = CLIContext(env={"edge_type": EdgeTypes.delete})
    assert PredecessorsPart.parse_args(None, w_delete) == (1, EdgeTypes.delete)
    assert PredecessorsPart.parse_args(None, plain) == (1, EdgeTypes.default)
    assert PredecessorsPart.parse_args("--with-origin", plain) == (0, EdgeTypes.default)
    assert PredecessorsPart.parse_args("--with-origin", w_delete) == (0, EdgeTypes.delete)
    assert PredecessorsPart.parse_args("--with-origin delete", plain) == (0, EdgeTypes.delete)
    assert PredecessorsPart.parse_args("--with-origin delete", w_delete) == (0, EdgeTypes.delete)
    assert PredecessorsPart.parse_args("delete", w_delete) == (1, EdgeTypes.delete)


@pytest.mark.asyncio
async def test_create_query_parts(cli: CLI) -> None:
    commands = await cli.evaluate_cli_command('search some_int==0 | search identifier=~"9_" | descendants')
    sort = "sort reported.kind asc, reported.name asc, reported.id asc"
    assert len(commands) == 1
    assert len(commands[0].commands) == 2  # list command is added automagically
    assert commands[0].commands[0].name == "execute_search"
    assert (
        commands[0].executable_commands[0].arg
        == f"'(reported.some_int == 0 and reported.identifier =~ \"9_\") {sort} -default[1:]-> all {sort}'"
    )
    commands = await cli.evaluate_cli_command("search some_int==0 | descendants")
    assert "-default[1:]->" in commands[0].executable_commands[0].arg  # type: ignore
    commands = await cli.evaluate_cli_command("search some_int==0 | ancestors | ancestors")
    assert "<-default[2:]-" in commands[0].executable_commands[0].arg  # type: ignore
    commands = await cli.evaluate_cli_command("search some_int==0 | predecessors | predecessors")
    assert "<-default[2]-" in commands[0].executable_commands[0].arg  # type: ignore
    commands = await cli.evaluate_cli_command("search some_int==0 | successors | successors | successors")
    assert "-default[3]->" in commands[0].executable_commands[0].arg  # type: ignore
    commands = await cli.evaluate_cli_command("search some_int==0 | successors | predecessors")
    assert f"-default-> all {sort} <-default-" in commands[0].executable_commands[0].arg  # type: ignore
    # defining the edge type is supported as well
    commands = await cli.evaluate_cli_command("search some_int==0 | successors delete")
    assert "-delete->" in commands[0].executable_commands[0].arg  # type: ignore
    commands = await cli.evaluate_cli_command("search some_int==0 | predecessors delete")
    assert "<-delete-" in commands[0].executable_commands[0].arg  # type: ignore
    commands = await cli.evaluate_cli_command("search some_int==0 | descendants delete")
    assert "-delete[1:]->" in commands[0].executable_commands[0].arg  # type: ignore
    commands = await cli.evaluate_cli_command("search some_int==0 | ancestors delete")
    assert "<-delete[1:]-" in commands[0].executable_commands[0].arg  # type: ignore
    commands = await cli.evaluate_cli_command("search some_int==0 | aggregate foo, bla as bla: sum(bar)")
    assert (
        commands[0].executable_commands[0].arg
        == f"'aggregate(reported.foo, reported.bla as bla: sum(reported.bar)):reported.some_int == 0 {sort}'"
    )

    # multiple head/tail commands are combined correctly
    commands = await cli.evaluate_cli_command("search is(volume) | head -10 | tail -5 | head -3")
    assert commands[0].executable_commands[0].arg == f"'is(\"volume\") {sort} limit 5, 3'"
    commands = await cli.evaluate_cli_command("search is(volume) sort name asc | head -10 | tail -5 | head -3")
    assert commands[0].executable_commands[0].arg == "'is(\"volume\") sort reported.name asc limit 5, 3'"
    commands = await cli.evaluate_cli_command("search is(volume) | head -10 | tail -5 | head -3 | tail 10 | head 100")
    assert commands[0].executable_commands[0].arg == f"'is(\"volume\") {sort} limit 5, 3'"
    commands = await cli.evaluate_cli_command("search is(volume) | tail -10")
    assert (
        commands[0].executable_commands[0].arg
        == f"'is(\"volume\") sort reported.kind desc, reported.name desc, reported.id desc limit 10 reversed '"
    )
    commands = await cli.evaluate_cli_command("search is(volume) sort name | tail -10 | head 5")
    assert commands[0].executable_commands[0].arg == "'is(\"volume\") sort reported.name desc limit 5, 5 reversed '"
    commands = await cli.evaluate_cli_command("search is(volume) sort name | tail -10 | head 5 | head 3 | tail 2")
    assert commands[0].executable_commands[0].arg == "'is(\"volume\") sort reported.name desc limit 7, 2 reversed '"


@pytest.mark.asyncio
async def test_replacements(cli: CLI) -> None:
    async def execute(template: str, replace_place_holder: bool = True) -> str:
        result = await cli.evaluate_cli_command(f"echo {template}", replace_place_holder=replace_place_holder)
        return result[0].parsed_commands.commands[0].args  # type: ignore

    # lookup keys are not case-sensitive
    today = utc().date().strftime("%Y-%m-%d")
    assert await execute("@today@ and @TODAY@") == f"{today} and {today}"

    # if the value is not provided, but a function is called
    assert await execute("@no_replacement@") == "@no_replacement@"

    # replacement is not touched if flag is set
    assert await execute("@today@", False) == "@today@"


def test_js_value_at() -> None:
    js = {"foo": {"bla": {"test": 123}}, "b": [{"a": 1, "b": [1, 2, 3]}, {"a": 2, "b": [1, 2, 3]}]}
    assert js_value_at(js, "b[0].a") == 1
    assert js_value_at(js, "b[0].b") == [1, 2, 3]
    assert js_value_at(js, "b[*].a") == [1, 2]
    assert js_value_at(js, "b[*].b[*]") == [[1, 2, 3], [1, 2, 3]]
    assert js_value_at(js, "b[*].b[2]") == [3, 3]
    assert js_value_at(js, "b[1].b[2]") == 3


def test_escape_character_parsing() -> None:
    def assert_command(to_parse: str, cmd: str, arg: str) -> None:
        parsed = multi_command_parser.parse(to_parse)
        assert len(parsed) == 1
        assert len(parsed[0].commands) == 1
        first = parsed[0].commands[0]
        assert first.cmd == cmd
        assert first.args == arg

    assert_command("echo 'f\\boo'", "echo", "'f\boo'")  # <bs>b --> \b
    assert_command("echo 'f\\noo'", "echo", "'f\noo'")  # <bs>n --> \n
    assert_command("echo 'f\\\\oo'", "echo", "'f\\\\oo'")  # <bs><bs> --> <bs><bs>
    assert_command("echo 'f\\u00a7oo'", "echo", "'f§oo'")  # <bs>unicode_number --> character
    assert_command("echo 'f\\\"oo'", "echo", "'f\\\"oo'")  # <bs>" --> <bs>"
    assert_command("echo 'f\\'oo'", "echo", "'f\\'oo'")  # <bs>' --> <bs>'


def test_get_kwargs() -> None:
    ec = ExecutableCommand(None, None, None, None)  # type: ignore
    d = {"a": 1, "b": "str", "previous_command": ec}
    assert CLICommand.get_previous_command(d) == ec
    assert CLICommand.get_from("a", int, d) == 1
    assert CLICommand.get_from("a", str, d) is None
    assert CLICommand.get_from("c", str, d) is None



from typing import Optional, List, Set, Any

from resotocore.model.model import Model, Kind
from resotocore.model.model_handler import ModelHandler
from resotocore.types import EdgeType
from resotocore.ids import GraphName


class ModelHandlerStatic(ModelHandler):
    def __init__(self, model: Model):
        self.model = model

    async def load_model(self, graph_name: GraphName, **kwargs: Any) -> Model:
        return self.model

    async def uml_image(
        self,
        graph_name: GraphName,
        output: str = "svg",
        *,
        show_packages: Optional[List[str]] = None,
        hide_packages: Optional[List[str]] = None,
        with_inheritance: bool = True,
        with_base_classes: bool = False,
        with_subclasses: bool = False,
        dependency_edges: Optional[Set[EdgeType]] = None,
        with_predecessors: bool = False,
        with_successors: bool = False,
        with_properties: bool = True,
        link_classes: bool = False,
        only_aggregate_roots: bool = True,
        sort_props: bool = True,
    ) -> bytes:
        raise NotImplementedError

    async def update_model(self, graph_name: GraphName, kinds: List[Kind]) -> Model:
        self.model = Model.from_kinds(kinds)
        return self.model



from typing import List

import pytest
from arango.database import StandardDatabase

from resotocore.config import ConfigEntity, ConfigValidation
from resotocore.db import configdb
from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.db.configdb import ConfigEntityDb, ConfigValidationEntityDb
from resotocore.ids import ConfigId


@pytest.fixture
async def config_db(test_db: StandardDatabase) -> ConfigEntityDb:
    async_db = AsyncArangoDB(test_db)
    cfg_db = configdb.config_entity_db(async_db, "configs")
    await cfg_db.create_update_schema()
    await cfg_db.wipe()
    return cfg_db


@pytest.fixture
async def validation_db(test_db: StandardDatabase) -> ConfigValidationEntityDb:
    async_db = AsyncArangoDB(test_db)
    cfg_db = configdb.config_validation_entity_db(async_db, "config_models")
    await cfg_db.create_update_schema()
    await cfg_db.wipe()
    return cfg_db


@pytest.fixture
def configs() -> List[ConfigEntity]:
    return [ConfigEntity(ConfigId(f"id_{a}"), {"some": a, "config": "test"}) for a in range(0, 10)]


@pytest.fixture
def config_models() -> List[ConfigValidation]:
    return [ConfigValidation(f"id_{a}", True) for a in range(0, 10)]


@pytest.mark.asyncio
async def test_load(config_db: ConfigEntityDb, configs: List[ConfigEntity]) -> None:
    await config_db.update_many(configs)
    loaded = [sub async for sub in config_db.all()]
    assert configs.sort() == loaded.sort()


@pytest.mark.asyncio
async def test_update(config_db: ConfigEntityDb, configs: List[ConfigEntity]) -> None:
    # multiple updates should work as expected
    await config_db.update_many(configs)
    await config_db.update_many(configs)
    await config_db.update_many(configs)
    loaded = [sub async for sub in config_db.all()]
    assert configs.sort() == loaded.sort()


@pytest.mark.asyncio
async def test_delete(config_db: ConfigEntityDb, configs: List[ConfigEntity]) -> None:
    await config_db.update_many(configs)
    remaining = list(configs)
    for _ in configs:
        sub = remaining.pop()
        await config_db.delete_value(sub)
        loaded = [sub async for sub in config_db.all()]
        assert remaining.sort() == loaded.sort()
    assert len([sub async for sub in config_db.all()]) == 0


@pytest.mark.asyncio
async def test_keys(config_db: ConfigEntityDb, configs: List[ConfigEntity]) -> None:
    assert [key async for key in config_db.keys()] == []
    await config_db.update_many(configs)
    assert [key async for key in config_db.keys()] == [a.id for a in configs]


@pytest.mark.asyncio
async def test_load_model(validation_db: ConfigValidationEntityDb, config_models: List[ConfigValidation]) -> None:
    await validation_db.update_many(config_models)
    loaded = [sub async for sub in validation_db.all()]
    assert config_models.sort() == loaded.sort()


@pytest.mark.asyncio
async def test_update_model(validation_db: ConfigValidationEntityDb, config_models: List[ConfigValidation]) -> None:
    # multiple updates should work as expected
    await validation_db.update_many(config_models)
    await validation_db.update_many(config_models)
    await validation_db.update_many(config_models)
    loaded = [sub async for sub in validation_db.all()]
    assert config_models.sort() == loaded.sort()


@pytest.mark.asyncio
async def test_delete_model(validation_db: ConfigValidationEntityDb, config_models: List[ConfigValidation]) -> None:
    await validation_db.update_many(config_models)
    remaining = list(config_models)
    for _ in config_models:
        sub = remaining.pop()
        await validation_db.delete_value(sub)
        loaded = [sub async for sub in validation_db.all()]
        assert remaining.sort() == loaded.sort()
    assert len([sub async for sub in validation_db.all()]) == 0


@pytest.mark.asyncio
async def test_keys_model(validation_db: ConfigValidationEntityDb, config_models: List[ConfigValidation]) -> None:
    assert [key async for key in validation_db.keys()] == []
    await validation_db.update_many(config_models)
    assert [key async for key in validation_db.keys()] == [a.id for a in config_models]



from concurrent.futures import ThreadPoolExecutor

from rich.text import Text

from resotocore.console_renderer import ConsoleColorSystem, ConsoleRenderer


def test_color_system() -> None:
    assert ConsoleColorSystem.monochrome.rich_color_system is None
    assert ConsoleColorSystem.eight_bit.rich_color_system == "256"
    assert ConsoleColorSystem.truecolor.rich_color_system == "truecolor"
    assert ConsoleColorSystem.legacy_windows.rich_color_system == "windows"


def test_from_name() -> None:
    for system in ConsoleColorSystem:
        assert ConsoleColorSystem.from_name(system.name) == system


def test_renderer_is_thread_safe() -> None:
    element = Text("some", "blue").append(Text("=", "dim").append("23", "green"))
    renderer = ConsoleRenderer(color_system=ConsoleColorSystem.standard)
    expected = renderer.render(element)

    def render() -> None:
        assert renderer.render(element) == expected

    with ThreadPoolExecutor(3) as pool:
        for n in range(0, 1000):
            pool.submit(render)

    assert len(ConsoleRenderer.console_pool.consoles) <= 3



from typing import Optional, List

from resotocore.cli import strip_quotes
from resotocore.cli.model import CLI, CLIContext
from resotocore.db.templatedb import TemplateEntityDb
from resotocore.query.model import Template
from resotocore.query.template_expander import TemplateExpanderBase
from resotocore.service import Service
from resotocore.types import Json


class TemplateExpanderService(TemplateExpanderBase, Service):
    """
    Template expander, which maintains the templates in the database.
    """

    def __init__(self, db: TemplateEntityDb, cli: CLI) -> None:
        self.db = db
        self.cli = cli

    def default_props(self) -> Optional[Json]:
        return None

    async def put_template(self, template: Template) -> None:
        await self.db.update(template)

    async def delete_template(self, name: str) -> None:
        await self.db.delete(name)

    async def get_template(self, name: str) -> Optional[Template]:
        return await self.db.get(name)

    async def list_templates(self) -> List[Template]:
        return [t async for t in self.db.all()]

    async def parse_query_from_command_line(self, to_parse: str, on_section: str, **env: str) -> str:
        final_env = {**env, "section": on_section, "no_rewrite": "true"}
        parsed = await self.cli.evaluate_cli_command(to_parse, CLIContext(env=final_env), replace_place_holder=False)
        if len(parsed) == 1:
            first_line = parsed[0]
            if len(first_line.executable_commands) == 1 and first_line.executable_commands[0].name == "execute_search":
                query = first_line.executable_commands[0].arg or ""
                return strip_quotes(query)
            else:
                iv = ", ".join(cmd.name for cmd in first_line.executable_commands[1:])
                raise ValueError(f"Commands found that can not be used as part of search: {iv}")
        else:
            raise ValueError("Multiple command lines are not supported in search!")



from typing import Callable, Dict, List

from attr import define

from resotocore.model.db_updater import GraphMerger
from resotocore.task.model import Subscriber


@define(eq=True, hash=True, order=True, repr=True, frozen=True)
class TaskDependencies:
    graph_merger: GraphMerger
    subscribers_by_event: Callable[[], Dict[str, List[Subscriber]]]



import asyncio
import logging
from datetime import timedelta
from typing import List

import pytest
from pytest import LogCaptureFixture

from resotocore.analytics import AnalyticsEventSender, InMemoryEventSender
from resotocore.cli.cli import CLIService
from tests.resotocore.conftest import eventually
from resotocore.db.jobdb import JobDb
from resotocore.db.runningtaskdb import RunningTaskDb
from resotocore.system_start import empty_config
from resotocore.ids import SubscriberId, TaskDescriptorId
from resotocore.message_bus import MessageBus, Event, Message, ActionDone, Action, ActionInfo, ActionError, CoreMessage
from resotocore.model.db_updater import GraphMerger
from resotocore.task.scheduler import Scheduler
from resotocore.task.subscribers import SubscriptionHandler
from resotocore.task.task_description import (
    Workflow,
    EventTrigger,
    TimeTrigger,
    Job,
    TaskSurpassBehaviour,
    ExecuteCommand,
    RunningTask,
)
from resotocore.task.task_handler import TaskHandlerService
from tests.resotocore.message_bus_test import wait_for_message


@pytest.mark.asyncio
async def test_run_job(task_handler: TaskHandlerService, all_events: List[Message]) -> None:
    await task_handler.handle_event(Event("start me up"))
    started: Event = await wait_for_message(all_events, "task_started", Event)
    await wait_for_message(all_events, "task_end", Event)
    assert started.data["task"] == "Speakable name of workflow"


@pytest.mark.asyncio
async def test_recover_workflow(
    running_task_db: RunningTaskDb,
    job_db: JobDb,
    message_bus: MessageBus,
    event_sender: AnalyticsEventSender,
    subscription_handler: SubscriptionHandler,
    graph_merger: GraphMerger,
    all_events: List[Message],
    cli: CLIService,
    test_workflow: Workflow,
) -> None:
    def handler() -> TaskHandlerService:
        th = TaskHandlerService(
            running_task_db,
            job_db,
            message_bus,
            event_sender,
            subscription_handler,
            graph_merger,
            Scheduler(),
            cli,
            empty_config(),
        )
        th.task_descriptions = [test_workflow]
        return th

    await subscription_handler.add_subscription(SubscriberId("sub_1"), "start_collect", True, timedelta(seconds=30))
    sub1 = await subscription_handler.add_subscription(SubscriberId("sub_1"), "collect", True, timedelta(seconds=30))
    sub2 = await subscription_handler.add_subscription(SubscriberId("sub_2"), "collect", True, timedelta(seconds=30))

    async with handler() as wf1:
        # kick off a new workflow
        await wf1.handle_event(Event("start me up"))
        assert len(wf1.tasks) == 1
        # expect a start_collect action message
        a: Action = await wait_for_message(all_events, "start_collect", Action)
        await wf1.handle_action_done(ActionDone(a.message_type, a.task_id, a.step_name, sub1.id, dict(a.data)))

        # expect a collect action message
        b: Action = await wait_for_message(all_events, "collect", Action)
        await wf1.handle_action_done(ActionDone(b.message_type, b.task_id, b.step_name, sub1.id, dict(b.data)))

    # subscriber 3 is also registering for collect
    # since the collect phase is already started, it should not participate in this round
    sub3 = await subscription_handler.add_subscription(SubscriberId("sub_3"), "collect", True, timedelta(seconds=30))

    # simulate a restart, wf1 is stopped and wf2 needs to recover from database
    async with handler() as wf2:
        assert len(wf2.tasks) == 1
        wfi = list(wf2.tasks.values())[0]
        assert wfi.current_state.name == "act"
        s1 = await wf2.list_all_pending_actions_for(sub1)
        s2 = await wf2.list_all_pending_actions_for(sub2)
        s3 = await wf2.list_all_pending_actions_for(sub3)
        assert (await wf2.list_all_pending_actions_for(sub1)) == []
        assert (await wf2.list_all_pending_actions_for(sub2)) == [Action("collect", wfi.id, "act", {})]
        assert (await wf2.list_all_pending_actions_for(sub3)) == []
        await wf2.handle_action_done(ActionDone("collect", wfi.id, "act", sub2.id, {}))
        # expect an event workflow_end
        await wait_for_message(all_events, "task_end", Event)
        # all workflow instances are gone
    assert len(wf2.tasks) == 0

    # simulate a restart, wf3 should start from a clean slate, since all instances are done
    async with handler() as wf3:
        assert len(wf3.tasks) == 0


@pytest.mark.asyncio
async def test_wait_for_running_job(
    task_handler: TaskHandlerService, test_workflow: Workflow, all_events: List[Message]
) -> None:
    test_workflow.on_surpass = TaskSurpassBehaviour.Wait
    task_handler.task_descriptions = [test_workflow]
    # subscribe as collect handler - the workflow will need to wait for this handler
    sub = await task_handler.subscription_handler.add_subscription(
        SubscriberId("sub_1"), "collect", True, timedelta(seconds=30)
    )
    await task_handler.handle_event(Event("start me up"))
    # check, that the workflow has started
    running_before = await task_handler.running_tasks()
    assert len(running_before) == 1
    act: Action = await wait_for_message(all_events, "collect", Action)
    # pull the same trigger: the workflow can not be started, since there is already one in progress -> wait
    await task_handler.handle_event(Event("start me up"))
    # report success of the only subscriber
    await task_handler.handle_action_done(ActionDone("collect", act.task_id, act.step_name, sub.id, dict(act.data)))
    # check overdue tasks: wipe finished tasks and eventually start waiting tasks
    await task_handler.check_running_tasks()
    # check, that the workflow has started
    running_after = await task_handler.running_tasks()
    assert len(running_after) == 1
    t_before, t_after = running_before[0], running_after[0]
    assert t_before.descriptor.id == t_after.descriptor.id and t_before.id != t_after.id


@pytest.mark.asyncio
async def test_handle_failing_task_command(task_handler: TaskHandlerService, caplog: LogCaptureFixture) -> None:
    caplog.set_level(logging.WARN)
    # This job will fail. Take a very long timeout - to avoid a timeout
    job = Job(TaskDescriptorId("fail"), ExecuteCommand("non_existing_command"), timedelta(hours=4))
    task_handler.task_descriptions = [job]
    assert len(await task_handler.running_tasks()) == 0
    await task_handler.start_task(job, "test fail")
    assert len(await task_handler.running_tasks()) == 1
    # The task is executed async - let's wait here directly
    update_task = (next(iter(task_handler.tasks.values()))).update_task
    assert update_task
    await update_task
    await task_handler.check_running_tasks()
    assert len(await task_handler.running_tasks()) == 0
    # One warning has been emitted
    assert len(caplog.records) == 1
    assert "Command non_existing_command failed with error" in caplog.records[0].message


@pytest.mark.asyncio
async def test_handle_failing_actor(
    task_handler: TaskHandlerService, test_workflow: Workflow, event_sender: InMemoryEventSender
) -> None:
    sub = await task_handler.subscription_handler.add_subscription(
        SubscriberId("sub_1"), "collect", True, timedelta(seconds=30)
    )
    info = await task_handler.start_task(test_workflow, "test")
    assert info is not None
    await asyncio.sleep(0.1)
    task = info.running_task
    await task_handler.handle_action_info(ActionInfo("collect", task.id, "act", sub.id, "error", "wrong!"))
    assert [e.kind for e in event_sender.events] == ["task-handler.task-started", "error.action"]
    await task_handler.handle_action_error(ActionError("collect", task.id, "act", sub.id, "wrong!"))
    assert [e.kind for e in event_sender.events] == ["task-handler.task-started", "error.action", "error.action"]


@pytest.mark.asyncio
async def test_default_workflow_triggers() -> None:
    workflows = {wf.name: wf for wf in TaskHandlerService.known_workflows(empty_config())}
    assert workflows["collect"].triggers == [EventTrigger("start_collect_workflow")]
    assert workflows["cleanup"].triggers == [EventTrigger("start_cleanup_workflow")]
    assert workflows["metrics"].triggers == [EventTrigger("start_metrics_workflow")]
    assert workflows["collect_and_cleanup"].triggers == [
        EventTrigger("start_collect_and_cleanup_workflow"),
        TimeTrigger("0 * * * *"),
    ]


@pytest.mark.asyncio
async def test_validate_add_delete_job(task_handler: TaskHandlerService) -> None:
    # add a job with a invalid name: should fail if the name contains invalid characters
    with pytest.raises(AttributeError):
        await task_handler.add_job(Job(TaskDescriptorId("foo:bar"), ExecuteCommand("echo foo"), timedelta(hours=4)))

    # add a job with a valid name: should succeed if the validation disabled
    await task_handler.add_job(
        Job(TaskDescriptorId("foo:bar"), ExecuteCommand("echo foo"), timedelta(hours=4)), force=True
    )

    # valid job name should not throw an exception
    await task_handler.add_job(Job(TaskDescriptorId("foobar"), ExecuteCommand("echo foo"), timedelta(hours=4)))

    # delete a job with a invalid name: should fail if the name contains invalid characters
    with pytest.raises(AttributeError):
        await task_handler.delete_job("foo:bar")

    # delete a job with a valid name: should succeed if the validation disabled
    await task_handler.delete_job("foobar", force=True)

    # force delete a job should not throw an exception
    await task_handler.delete_job("foo:bar", force=True)


@pytest.mark.asyncio
async def test_wait_for_collect_done(
    task_handler: TaskHandlerService, message_bus: MessageBus, graph_merger: GraphMerger
) -> None:
    async with task_handler:
        # Test 1: start a task without any ongoing graph merges
        # start task
        task = await task_handler.start_task_by_descriptor_id(TaskDescriptorId("wait_for_collect_done"))
        assert task is not None
        rt: RunningTask = task.running_task
        assert rt in await task_handler.running_tasks()
        # no collect is in progress. The task should be finished immediately.
        await message_bus.emit_event("collected", {})
        await eventually(lambda: rt.current_step.name == "task_end")

        # Test 2: start a task with an ongoing graph merge operations. The task should wait for the merge to finish.
        # start task
        task = await task_handler.start_task_by_descriptor_id(TaskDescriptorId("wait_for_collect_done"))
        assert task is not None
        rt = task.running_task
        assert rt in await task_handler.running_tasks()
        # fake some ongoing imports in graph merger
        graph_merger.running_imports[rt.id] = 42
        # send event to finish task
        await message_bus.emit_event("collected", {})
        # sleep a little bit to make sure the task is not finished
        await asyncio.sleep(0.1)
        await eventually(lambda: rt.current_step.name == "wait_for_collect_done")
        # signal, that the import is finished
        await message_bus.emit_event(CoreMessage.GraphMergeCompleted, dict(task_id=rt.id))
        # make sure the task finishes
        await eventually(lambda: rt.current_step.name == "task_end")



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from enum import Enum
from functools import reduce
from typing import List, Optional, Dict, ClassVar, AsyncIterator, cast, Set, Tuple

from attr import define, field, evolve

from resotocore.ids import ConfigId, GraphName
from resotocore.model.typed_model import to_js
from resotocore.types import Json
from resotocore.util import uuid_str, if_set, partition_by

log = logging.getLogger(__name__)

# config ids
ResotoReportValues = ConfigId("resoto.report.values")
ResotoReportBenchmark = ConfigId("resoto.report.benchmark")
ResotoReportCheck = ConfigId("resoto.report.check")

# config keys and prefixes
CheckConfigPrefix = "resoto.report.check."
CheckConfigRoot = "report_check"
BenchmarkConfigPrefix = "resoto.report.benchmark."
BenchmarkConfigRoot = "report_benchmark"


class ReportSeverity(Enum):
    kind: ClassVar[str] = "resoto_core_report_check_severity"
    info = "info"
    low = "low"
    medium = "medium"
    high = "high"
    critical = "critical"


ReportSeverityPriority: Dict[ReportSeverity, int] = {severity: num for num, severity in enumerate(ReportSeverity)}


@define
class Remediation:
    kind: ClassVar[str] = "resoto_core_report_check_remediation"
    text: str = field(metadata={"description": "Textual description of the remediation."})
    url: str = field(metadata={"description": "URL that documents the remediation action."})
    action: Optional[Dict[str, str]] = field(
        default=None,
        metadata={
            "description": "Remediation actions with tool as key.\n"
            "Example tools: resoto_cmd, awscli, gcloud, terraform"
        },
    )


@define
class ReportCheck:
    id: str
    provider: str
    service: str
    title: str
    result_kind: str
    categories: List[str]
    severity: ReportSeverity
    risk: str
    detect: Dict[str, str]
    remediation: Remediation
    default_values: Optional[Json] = None
    url: Optional[str] = None
    related: Optional[List[str]] = None

    def environment(self, values: Json) -> Json:
        return {**self.default_values, **values} if self.default_values else values

    def to_node(self) -> Json:
        reported = to_js(self)
        return dict(id=self.id, kind="report_check", type="node", reported=reported)


@define
class CheckCollection:
    title: str
    description: str
    documentation: Optional[str]
    checks: Optional[List[str]]
    children: Optional[List[CheckCollection]]

    def is_valid(self, checks: Dict[str, ReportCheck]) -> bool:
        return (
            not (self.children and self.checks)
            and all(c in checks for c in self.checks or [])  # pylint: disable=not-an-iterable
            and all(c.is_valid(checks) for c in self.children or [])  # pylint: disable=not-an-iterable
        )

    def nested_checks(self) -> List[str]:
        result: List[str] = []
        if self.checks:
            result.extend(c for c in self.checks)  # pylint: disable=not-an-iterable
        if self.children:
            for child in self.children:  # pylint: disable=not-an-iterable
                result.extend(child.nested_checks())
        return result


@define
class Benchmark(CheckCollection):
    id: str
    framework: str
    version: str
    clouds: Optional[List[str]] = None

    def to_node(self) -> Json:
        reported = to_js(self)
        return dict(id=self.id, kind="benchmark", type="node", reported=reported)


@define
class CheckResult:
    check: ReportCheck
    number_of_resources_failing_by_account: Dict[str, int]
    node_id: str = field(init=False, factory=uuid_str)

    @property
    def number_of_resources_failing(self) -> int:
        return reduce(lambda a, b: a + b, self.number_of_resources_failing_by_account.values(), 0)

    def has_failed(self) -> bool:
        return self.number_of_resources_failing > 0

    def to_node(self) -> Json:
        reported = to_js(self.check, strip_attr="kind")
        reported["kind"] = "report_check_result"
        reported["name"] = self.check.title
        reported["number_of_resources_failing"] = self.number_of_resources_failing
        if self.number_of_resources_failing_by_account:
            reported["number_of_resources_failing_by_account"] = self.number_of_resources_failing_by_account
        return dict(id=self.node_id, kind="report_check_result", type="node", reported=reported)

    @staticmethod
    def from_node(js: Json) -> CheckResult:
        reported = cast(Json, js["reported"])
        return CheckResult(
            check=ReportCheck(
                id=reported["id"],
                provider=reported["provider"],
                service=reported["service"],
                title=reported["title"],
                result_kind=reported["result_kind"],
                categories=reported["categories"],
                severity=ReportSeverity(reported["severity"]),
                risk=reported["risk"],
                detect=reported["detect"],
                remediation=Remediation(**reported["remediation"]),
                default_values=reported.get("default_values"),
                url=reported.get("url"),
                related=reported.get("related", []),
            ),
            number_of_resources_failing_by_account=reported.get("number_of_resources_failing_by_account", {}),
        )


@define
class CheckCollectionResult:
    title: str
    description: str
    documentation: Optional[str] = field(default=None, kw_only=True)
    checks: List[CheckResult] = field(factory=list, kw_only=True)
    children: List[CheckCollectionResult] = field(factory=list, kw_only=True)
    node_id: str = field(init=False, factory=uuid_str)

    def to_node(self) -> Json:
        return dict(
            id=self.node_id,
            type="node",
            reported=dict(
                kind="report_check_collection",
                name=self.title,
                title=self.title,
                description=self.description,
                documentation=self.documentation,
            ),
        )

    @staticmethod
    def from_node(js: Json) -> CheckCollectionResult:
        reported = cast(Json, js["reported"])
        return CheckCollectionResult(
            title=reported["title"],
            description=reported["description"],
            documentation=reported.get("documentation"),
        )

    def is_empty(self) -> bool:
        return not self.checks and not self.children

    def has_failed(self) -> bool:
        return any(c.has_failed() for c in self.checks) or any(c.has_failed() for c in self.children)

    def has_failed_for_account(self, account: str) -> bool:
        return any(account in c.number_of_resources_failing_by_account for c in self.checks) or any(
            c.has_failed_for_account(account) for c in self.children
        )

    def passing_failing_checks_for_account(self, account: str) -> Tuple[List[CheckResult], List[CheckResult]]:
        passing_child_checks, failing_child_checks = reduce(
            lambda pf, el: (pf[0] + el[0], pf[1] + el[1]),
            [c.passing_failing_checks_for_account(account) for c in self.children],
            (cast(List[CheckResult], []), cast(List[CheckResult], [])),
        )

        failing_checks, passing_checks = partition_by(
            lambda c: account in c.number_of_resources_failing_by_account, self.checks
        )
        return passing_checks + passing_child_checks, failing_checks + failing_child_checks

    def passing_failing_checks_count_for_account(self, account: str) -> Tuple[int, int]:
        passing, failing = reduce(
            lambda pf, el: (pf[0] + el[0], pf[1] + el[1]),
            [c.passing_failing_checks_count_for_account(account) for c in self.children],
            (0, 0),
        )

        all_checks = len(self.checks)
        failing_count = sum(1 for c in self.checks if account in c.number_of_resources_failing_by_account)
        return passing + all_checks - failing_count, failing + failing_count

    def filter_result(
        self, filter_failed: bool = False, failed_for_account: Optional[str] = None
    ) -> CheckCollectionResult:
        return evolve(
            self,
            checks=[
                c
                for c in self.checks
                if (not filter_failed or c.has_failed())
                and (failed_for_account is None or failed_for_account in c.number_of_resources_failing_by_account)
            ],
            children=[
                c.filter_result(filter_failed, failed_for_account)
                for c in self.children
                if (not filter_failed or c.has_failed())
                and (c.checks or c.children)
                and (failed_for_account is None or c.has_failed_for_account(failed_for_account))
            ],
        )

    def check_results(self) -> List[CheckResult]:
        return self.checks + [c for child in self.children for c in child.check_results()]

    def failing_accounts(self) -> Set[str]:
        return {account for result in self.check_results() for account in result.number_of_resources_failing_by_account}


@define
class BenchmarkResult(CheckCollectionResult):
    framework: str
    version: str
    accounts: Optional[List[str]] = field(default=None)
    only_failed: bool = field(default=False)
    severity: Optional[ReportSeverity] = field(default=None)

    def to_node(self) -> Json:
        node = super().to_node()
        reported = node["reported"]
        reported["framework"] = self.framework
        reported["version"] = self.version
        reported["kind"] = "report_benchmark"
        reported["accounts"] = self.accounts
        reported["only_failed"] = self.only_failed
        reported["severity"] = if_set(self.severity, lambda s: s.value)
        return node

    @staticmethod
    def from_node(js: Json) -> BenchmarkResult:
        reported = cast(Json, js["reported"])
        return BenchmarkResult(
            framework=reported["framework"],
            version=reported["version"],
            title=reported["title"],
            description=reported["description"],
            documentation=reported.get("documentation"),
            accounts=reported["accounts"],
            only_failed=reported["only_failed"],
            severity=if_set(reported.get("severity"), ReportSeverity),
        )

    def to_graph(self, only_checks: bool = False) -> List[Json]:
        result = []

        def visit_check_collection(collection: CheckCollectionResult) -> None:
            if not only_checks:
                result.append(collection.to_node())
            for check in collection.checks:
                result.append(check.to_node())
                if not only_checks:
                    result.append(
                        {"from": collection.node_id, "to": check.node_id, "type": "edge", "edge_type": "default"}
                    )
            for child in collection.children:
                visit_check_collection(child)
                if not only_checks:
                    result.append(
                        {"from": collection.node_id, "to": child.node_id, "type": "edge", "edge_type": "default"}
                    )

        visit_check_collection(self)
        return result


class Inspector(ABC):
    """
    The inspector is able to perform benchmarks or a set of checks.
    Benchmarks and checks are maintained via the configuration registry.
    A complete graph can be checked against either a pre-defined benchmark or from a selected set of checks.
    """

    @abstractmethod
    async def list_benchmarks(self) -> List[Benchmark]:
        pass

    @abstractmethod
    async def benchmark(self, name: str) -> Optional[Benchmark]:
        pass

    @abstractmethod
    async def list_checks(
        self,
        *,
        provider: Optional[str] = None,
        service: Optional[str] = None,
        category: Optional[str] = None,
        kind: Optional[str] = None,
        check_ids: Optional[List[str]] = None,
    ) -> List[ReportCheck]:
        """
        List all inspection checks matching the given criteria.
        If no criteria is given, all checks are returned.

        :param provider: the provider of the check (e.g. aws, gcp, k8s, ...)
        :param service: the service inside the provider (e.g. ec2, lambda, ...)
        :param category: the category of the check (e.g. security, compliance, cost ...)
        :param kind: the resulting kind of the check (e.g. aws_ec2_instance, kubernetes_pod, ...)
        :param check_ids: the list of check ids to return
        :return: the list of matching checks
        """

    @abstractmethod
    async def perform_benchmark(
        self,
        graph: GraphName,
        benchmark_name: str,
        *,
        accounts: Optional[List[str]] = None,
        severity: Optional[ReportSeverity] = None,
        only_failing: bool = False,
    ) -> BenchmarkResult:
        """
        Perform a benchmark by given name on the content of a graph with given name.

        :param benchmark_name: the name of the benchmark to perform (e.g. aws_cis_1_5_0)
        :param graph: the name of the graph to perform the benchmark on (e.g. resoto)
        :param accounts: the list of accounts to perform the benchmark on. If not given, all accounts are used.
        :param severity: only include checks with given severity or higher
        :param only_failing: only include failing checks in the result
        :return: the result of the benchmark
        """

    @abstractmethod
    async def perform_checks(
        self,
        graph: GraphName,
        *,
        provider: Optional[str] = None,
        service: Optional[str] = None,
        category: Optional[str] = None,
        kind: Optional[str] = None,
        check_ids: Optional[List[str]] = None,
        accounts: Optional[List[str]] = None,
        severity: Optional[ReportSeverity] = None,
        only_failing: bool = False,
    ) -> BenchmarkResult:
        """
        Perform a benchmark by selecting all checks matching the given criteria.

        :param graph: the name of the graph to perform the checks on (e.g. resoto)
        :param provider: the provider of the check (e.g. aws, gcp, k8s, ...)
        :param service: the service inside the provider (e.g. ec2, lambda, ...)
        :param category: the category of the check (e.g. security, compliance, cost ...)
        :param kind: the resulting kind of the check (e.g. aws_ec2_instance, kubernetes_pod, ...)
        :param check_ids: the ids of the checks to perform.
        :param accounts: the list of accounts to perform the benchmark on. If not given, all accounts are used.
        :param severity: only include checks with given severity or higher
        :param only_failing: only include failing checks in the result
        :return: the result of this benchmark
        """

    @abstractmethod
    async def list_failing_resources(
        self, graph: GraphName, check_uid: str, account_ids: Optional[List[str]] = None
    ) -> AsyncIterator[Json]:
        pass

    @abstractmethod
    async def validate_benchmark_config(self, json: Json) -> Optional[Json]:
        pass

    @abstractmethod
    async def validate_check_collection_config(self, json: Json) -> Optional[Json]:
        pass



import logging
import os
import re
from argparse import Namespace
from contextlib import suppress
from copy import deepcopy
from datetime import timedelta
from pathlib import Path
from typing import Optional, List, ClassVar, Dict, Union, cast, Callable

from arango.database import StandardDatabase
from attrs import define, field
from cerberus import schema_registry

from resotocore.ids import ConfigId
from resotocore.model.model import Kind, Model, ComplexKind
from resotocore.model.typed_model import from_js, to_js
from resotocore.types import Json, JsonElement
from resotocore.util import set_value_in_path, value_in_path, del_value_in_path
from resotocore.validator import Validator, schema_name
from resotolib.core.model_export import dataclasses_to_resotocore_model
from resotolib.utils import replace_env_vars, is_env_var_string, merge_json_elements

log = logging.getLogger(__name__)

# ids used in the config store
ResotoCoreConfigId = ConfigId("resoto.core")
ResotoCoreCommandsConfigId = ConfigId("resoto.core.commands")
ResotoCoreSnapshotsConfigId = ConfigId("resoto.core.snapshots")

# root note of the configuration value
ResotoCoreRoot = "resotocore"
ResotoCoreCommandsRoot = "custom_commands"
ResotoCoreSnapshotsRoot = "snapshots"

ResotoCoreRootRE = re.compile(r"^resotocore[.]")

# created by the docker build process
GitHashFile = "/usr/local/etc/git-commit.HEAD"


def git_hash_from_file() -> Optional[str]:
    """
    Returns the git hash from the file created by the docker build.
    In case we do not run inside a docker container, this method returns None.
    """
    with suppress(Exception):
        path = Path(GitHashFile)
        if path.exists():
            return path.read_text("utf-8").strip()
    return None


def inside_docker() -> bool:
    """
    Try to detect if we are running inside a docker container.
    """
    return (
        # environment variables have to be set explicitly
        os.environ.get("INSIDE_DOCKER", "false").lower() in ("true", "yes", "1")
        or os.environ.get("INSIDE_KUBERNETES", "false").lower() in ("true", "yes", "1")
        # this file is available in the created docker container
        or git_hash_from_file() is not None
    )


def inside_kubernetes() -> bool:
    """
    Try to detect if we are running on kubernetes.
    """
    # environment variables have to be set explicitly
    return "HELM_VERSION" in os.environ or any(True for x in os.environ if x.startswith("KUBERNETES_"))


def helm_installation() -> bool:
    """
    Try to detect if we were installed via helm chart.
    """
    # environment variables have to be set explicitly
    return "HELM_VERSION" in os.environ


def default_hosts() -> List[str]:
    return ["0.0.0.0"] if inside_docker() else ["localhost"]


def validate_config(config: Json, clazz: type) -> Optional[Json]:
    def strip_env_vars_paths(config: JsonElement) -> JsonElement:
        """
        Recursively strips all values that contain an env var string
        """
        if isinstance(config, dict):
            return {k: strip_env_vars_paths(v) for k, v in config.items() if not is_env_var_string(v)}
        elif isinstance(config, list):
            return [strip_env_vars_paths(v) for v in config if not is_env_var_string(v)]
        else:
            return config

    schema = schema_name(clazz)
    v = Validator(schema=schema, allow_unknown=True)
    # cerberus is too inflexible to allow us to validate the config without resolving the env vars
    # so we have to strip strings with the env vars before validating
    without_env_vars = strip_env_vars_paths(config)
    result = v.validate(without_env_vars, normalize=False)
    return None if result else v.errors


class ConfigObject:
    def validate(self) -> Optional[Json]:
        return validate_config(to_js(self), type(self))


@define()
class CertificateConfig(ConfigObject):
    kind: ClassVar[str] = f"{ResotoCoreRoot}_certificate_config"
    common_name: str = field(default="some.engineering", metadata={"description": "The common name of the certificate"})
    include_loopback: bool = field(default=True, metadata={"description": "Include loopback in certificate"})
    san_dns_names: List[str] = field(factory=list, metadata={"description": "List of DNS names to include in CSR"})
    san_ip_addresses: List[str] = field(
        factory=list, metadata={"description": "List of IP addresses to include in CSR"}
    )


@define()
class ApiConfig(ConfigObject):
    kind: ClassVar[str] = f"{ResotoCoreRoot}_api_config"

    web_hosts: List[str] = field(
        factory=default_hosts, metadata={"description": f"TCP host(s) to bind on (default: {default_hosts()})"}
    )
    https_port: Optional[int] = field(
        default=8900, metadata={"description": "TCP port to bind on for TLS encrypted connections (default: 8900)"}
    )
    http_port: Optional[int] = field(
        default=8980, metadata={"description": "TCP port to bind on for plain HTTP connections (default: 8980)"}
    )
    web_path: str = field(
        default="/",
        metadata={
            "description": "Web path root (default: /).\n"
            "This should only be required, if you are running a proxy server, that is not able to handle a sub-path."
        },
    )
    tsdb_proxy_url: Optional[str] = field(
        default=None,
        metadata={"description": "The url to the time series database. This path will be served under /tsdb/."},
    )
    max_request_size: Optional[int] = field(
        default=1024**2 * 5, metadata={"description": "The maximum size of a request in bytes (default: 5MB)"}
    )
    host_certificate: CertificateConfig = field(
        factory=CertificateConfig, metadata={"description": "The certificate configuration for this server."}
    )
    access_token_expiration_seconds: int = field(
        default=3600, metadata={"description": "The expiration time of the access token in seconds (default: 1h)"}
    )

    def access_token_expiration(self) -> timedelta:
        return timedelta(seconds=self.access_token_expiration_seconds)


# Define rules to validate this config
schema_registry.add(
    schema_name(ApiConfig),
    dict(
        http_port={"type": "integer", "min": 1, "max": 65535, "nullable": True},
        https_port={"type": "integer", "min": 1, "max": 65535, "nullable": True},
        tsdb_proxy_url={"type": "string", "nullable": True, "is_url": True},
        max_request_size={"type": "integer", "nullable": True, "min": 1024**2},
    ),
)


@define()
class DatabaseConfig(ConfigObject):
    kind: ClassVar[str] = f"{ResotoCoreRoot}_database_config"
    server: str = field(
        default="http://localhost:8529",
        metadata={"description": "Graph database server (default: http://localhost:8529)"},
    )
    database: str = field(default="resoto", metadata={"description": "Graph database name (default: resoto)"})
    username: str = field(default="resoto", metadata={"description": "Graph database login (default: resoto)"})
    password: str = field(default="", metadata={"description": 'Graph database password (default: "")'})
    root_password: str = field(
        default="",
        metadata={"description": "Graph root database password used for creating user and database if not existent."},
    )
    bootstrap_do_not_secure: bool = field(
        default=False, metadata={"description": "Leave an empty root password during system setup process."}
    )
    no_ssl_verify: bool = field(
        default=False, metadata={"description": "If the connection should not be verified (default: False)"}
    )
    request_timeout: int = field(default=900, metadata={"description": "Request timeout in seconds (default: 900)"})


@define(order=True, hash=True, frozen=True)
class AliasTemplateParameterConfig(ConfigObject):
    kind: ClassVar[str] = f"{ResotoCoreRoot}_cli_alias_template_parameter"
    name: str = field(metadata=dict(description="The name of the parameter."))
    description: str = field(metadata=dict(description="The intent of this parameter."))
    default: Optional[JsonElement] = field(
        default=None,
        metadata=dict(
            description="The optional default value.\n"
            "In case a default value exists, it does not need to be provided by the user."
        ),
    )


@define(order=True, hash=True, frozen=True)
class AliasTemplateConfig(ConfigObject):
    kind: ClassVar[str] = f"{ResotoCoreRoot}_cli_alias_template"
    name: str = field(metadata=dict(description="The name of the alias to execute."))
    info: str = field(metadata=dict(description="A one line sentence that describes the effect of this command."))
    template: str = field(metadata=dict(description="The command to execute which can have template parameters."))
    parameters: List[AliasTemplateParameterConfig] = field(
        factory=list, metadata=dict(description="All template parameters.")
    )
    description: Optional[str] = field(metadata=dict(description="A longer description of the command."), default=None)
    allowed_in_source_position: Optional[bool] = field(
        metadata=dict(
            description="true if this alias can be executed directly, false if it expects input from another command."
        ),
        default=False,
    )


def alias_templates() -> List[AliasTemplateConfig]:
    return [
        AliasTemplateConfig(
            name="jira",
            info="Send the result of a search to Jira",
            description=(
                "Perform a search and send the result to Jira.\n\n"
                "If your search result is larger than 25 items, only the first 25 items will be added to the ticket, "
                "and the remaining items will be dropped.\n\n"
                "Note that invoking this command will always create a new ticket since JIRA does not have any "
                "deduplication functionality.\n\n"
                "We recommend to define the URL, username and token as part of the command configuration. "
                "This way you do not need to provide it every time you execute the command."
            ),
            template=(
                # defines the fields to show in the message
                'head 26 | jq ({{key}} + ": " + {{value}}) | chunk 26 | '
                'jq \'((.[:25] | join("\\n")) + (if .[25] then "\\n... (results truncated)" else "" end))\' | '
                # define the Jira webhook json
                "jq {fields: { "
                'summary: "{{title}}", '
                'issuetype: {id: "10001"}, '
                'description: ("{{message}}" + "\\n\\n" + . + "\\n\\n" + "Issue created by Resoto"), '
                'project: {id: "{{project_id}}"}, '
                'reporter: {id: "{{reporter_id}}"}, '
                'labels: ["created-by-resoto"]'
                "}}"
                # call the api
                '| http --auth "{{username}}:{{token}}" POST {{url}}/rest/api/2/issue'
            ),
            parameters=[
                AliasTemplateParameterConfig("key", "Resource field to show as key", ".kind"),
                AliasTemplateParameterConfig("value", "Resource field to show as value", ".name"),
                AliasTemplateParameterConfig("message", "Alert message", ""),
                AliasTemplateParameterConfig("title", "Alert title"),
                AliasTemplateParameterConfig("url", "Jira URL"),
                AliasTemplateParameterConfig("username", "Jira username"),
                AliasTemplateParameterConfig("token", "Jira API token"),
                AliasTemplateParameterConfig("project_id", "Jira project ID"),
                AliasTemplateParameterConfig("reporter_id", "Jira reporter user ID"),
            ],
            allowed_in_source_position=False,
        ),
        AliasTemplateConfig(
            name="alertmanager",
            info="Create an alert in alertmanager from a search.",
            description=(
                "Perform a search and send the result to alertmanager.\n\n"
                "No resource specific data will be sent to alertmanager - only the count of matching resources. "
                "The alert will be created in alertmanager and will be active for the specified duration.\n\n"
                "The name of the alert is visible in alertmanager and used as deduplication key. "
                "This way the same alert can be fired multiple times.\n\n"
                "We recommend to define the URL as part of the command configuration. "
                "This way you do not need to provide it every time you execute the command."
            ),
            template=(
                "aggregate sum(1) as count | "
                # do not send an alert in case of 0 violations
                'jq --no-rewrite "if (.count // 0)==0 then [] else [.count | tostring] end" | flatten | '
                # defines the fields to show in the message
                "jq --no-rewrite [{"
                'status: "firing", '
                'labels: {alertname: "{{name}}", issued_by: "Resoto"}, '
                'annotations: {summary: ("Found "+.+ " violations!"), '
                '"description": "{{description}}"}{{#duration}}, '
                'startAt:"@utc@", '
                'endsAt:"{{duration.from_now}}"{{/duration}}}] | '
                # call the api
                "http POST {{alertmanager_url}}/api/v1/alerts"
            ),
            parameters=[
                AliasTemplateParameterConfig("name", "The globally unique name of this alert."),
                AliasTemplateParameterConfig("description", "User defined message of the post.", "Resoto Alert"),
                AliasTemplateParameterConfig("duration", "The duration of this alert in alertmanager.", "3h"),
                AliasTemplateParameterConfig("alertmanager_url", "The complete url to alertmanager."),
            ],
            allowed_in_source_position=False,
        ),
        AliasTemplateConfig(
            name="pagerduty",
            info="Create an alert in pagerduty from a search.",
            description=(
                "Perform a search and send the result to pagerduty.\n\n"
                "A call to this command will only send the first 100 occurrences to the incident, the rest is dropped. "
                "The `summary` should explain why this alert is triggered, so that the user can take action.\n"
                "The `dedup_key` is used to identify an alert uniquely. "
                "You can fire the same alert multiple times by using the same dedup_key.\n\n"
                "We recommend to define the `routing_key` as part of the command configuration. "
                "This way you do not need to provide it every time you execute the command."
            ),
            template=(
                # aggregate the result by cloud -> account -> region -> resource
                # resulting structure looks like this:
                # {"aws": {"account1": {"region1": {"id1": {"id": "xxx", "name": "yyy", "kind": "zzz" }}}}}
                # note: Pagerduty is able to render JSON objects in their webUI, but not arrays.
                "head 100 | chunk 100 | jq --no-rewrite '"
                "{{#group_resources}}"
                '[group_by(.ancestors.cloud.reported.name) | .[] | {(.[0].ancestors.cloud.reported.name // "no-cloud"): '  # noqa: E501
                '[group_by(.ancestors.account.reported.name) | .[] | {(.[0].ancestors.account.reported.name // "no-account"): '  # noqa: E501
                '[group_by(.ancestors.region.reported.name) | .[] | {(.[0].ancestors.region.reported.name // "no-region"): '  # noqa: E501
                "{{/group_resources}}"
                "[.[] | {({{resource_id}}): { {{#resource_properties.as_list.with_index}}{{key}}: {{value}}{{^last}},{{/last}}{{/resource_properties.as_list.with_index}} }}] | add "  # noqa: E501
                "{{#group_resources}}}] | add }] | add }] | add {{/group_resources}}'"
                "| jq --no-rewrite '{payload: "
                '{summary: "{{summary}}", '
                'timestamp: "@utc@", '
                'source:"{{source}}", '
                'severity: "{{severity}}", '
                'component: "{{component}}", '
                "custom_details: .}, "
                'routing_key: "{{routing_key}}", '
                'dedup_key: "{{dedup_key}}", '
                'images:[{src: "https://cdn.some.engineering/assets/resoto-illustrations/small/resoto-alert.png", href:'
                ' "https://resoto.com/", alt: "Resoto Home Page"}], '
                "links:[], "
                'event_action: "{{event_action}}", '
                'client: "Resoto Service", '
                'client_url: "https://resoto.com"}\''
                # send the event to pagerduty
                ' | http {{webhook_url}} "Content-Type:application/json"'
            ),
            parameters=[
                AliasTemplateParameterConfig("summary", "The summary of this alert."),
                AliasTemplateParameterConfig(
                    "severity",
                    "The perceived severity of the status the event is describing withrespect to the affected system. "
                    "One of: `critical`, `error`, `warning` or `info`.",
                    "warning",
                ),
                AliasTemplateParameterConfig(
                    "source", "The unique location of the affected system, preferably a hostname or FQDN.", "Resoto"
                ),
                AliasTemplateParameterConfig(
                    "component", "Component of the source machine that is responsible for the event.", "Resoto"
                ),
                AliasTemplateParameterConfig(
                    "routing_key",
                    "The GUID of one of your Events API V2 integrations. "
                    'This is the "Integration Key" listed on the Events API V2 integration\'s detail page.',
                ),
                AliasTemplateParameterConfig(
                    "event_action", "The type of event. Can be `trigger`, `acknowledge` or `resolve`.", "trigger"
                ),
                AliasTemplateParameterConfig("dedup_key", "Identifies the alert to trigger."),
                AliasTemplateParameterConfig(
                    "client", "The name of the monitoring client that is triggering this event.", "Resoto"
                ),
                AliasTemplateParameterConfig(
                    "client_url",
                    "The URL of the monitoring client that is triggering this event.",
                    "https://resoto.com",
                ),
                AliasTemplateParameterConfig(
                    "webhook_url",
                    "The complete url of the pagerduty events API.",
                    "https://events.pagerduty.com/v2/enqueue",
                ),
                AliasTemplateParameterConfig(
                    "group_resources",
                    "Group Resource by cloud, account, and region.",
                    True,
                ),
                AliasTemplateParameterConfig(
                    "resource_id",
                    "Property to show as resource identifier.",
                    ".id",
                ),
                AliasTemplateParameterConfig(
                    "resource_properties",
                    "Dictionary of properties to show.",
                    dict(id=".reported.id", name=".reported.name", kind=".reported.kind"),
                ),
            ],
            allowed_in_source_position=False,
        ),
    ]


@define()
class CLIConfig(ConfigObject):
    kind: ClassVar[str] = f"{ResotoCoreRoot}_cli_config"
    default_graph: str = field(
        default="resoto",
        metadata={"description": "Use this graph for CLI actions, if no graph is specified explicitly."},
    )
    default_section: str = field(
        default="reported",
        metadata={
            "description": "Use this graph section by default, if no section is specified.\n"
            "Relative paths will be interpreted with respect to this section."
        },
    )


# Define rules to validate this config
schema_registry.add(schema_name(CLIConfig), {})


@define()
class CustomCommandsConfig(ConfigObject):
    kind: ClassVar[str] = ResotoCoreCommandsRoot
    commands: List[AliasTemplateConfig] = field(
        factory=alias_templates,
        metadata={"description": "Here you can define all custom commands for the CLI."},
    )

    def json(self) -> Json:
        return {ResotoCoreCommandsRoot: to_js(self, strip_attr="kind")}


# Define rules to validate this config
schema_registry.add(schema_name(CustomCommandsConfig), {})


SnapshotLabel = str


@define
class SnapshotSchedule(ConfigObject):
    kind: ClassVar[str] = f"{ResotoCoreSnapshotsRoot}_schedule"
    schedule: str = field(
        metadata={
            "description": "The schedule in cron format.\n"
            "Example: `0 0 * * *` will create a snapshot every day at midnight.\n"
            "See https://en.wikipedia.org/wiki/Cron for more information.",
        }
    )
    retain: int = field(
        metadata={
            "description": "How many snapshots should be retained.\n"
            "If the number of snapshots exceeds this value, the oldest snapshots will be deleted.\n"
        }
    )


@define()
class SnapshotsScheduleConfig(ConfigObject):
    kind: ClassVar[str] = ResotoCoreSnapshotsRoot
    snapshots: Dict[SnapshotLabel, SnapshotSchedule] = field(
        default={
            "hourly": SnapshotSchedule(schedule="0 * * * *", retain=24),
            "daily": SnapshotSchedule(schedule="0 0 * * *", retain=7),
            "weekly": SnapshotSchedule(schedule="0 0 * * 0", retain=4),
            "monthly": SnapshotSchedule(schedule="0 0 1 * *", retain=12),
            "yearly": SnapshotSchedule(schedule="0 0 1 1 *", retain=10),
        },
        metadata={
            "description": "Here you can define all snapshot schedules.\n"
            "The key is the label of the snapshot schedule.\n"
            "The value is the schedule configuration.",
        },
    )

    def json(self) -> Json:
        return to_js(self, strip_attr="kind")


@define()
class GraphUpdateConfig(ConfigObject):
    kind: ClassVar[str] = f"{ResotoCoreRoot}_graph_update_config"
    merge_max_wait_time_seconds: int = field(
        default=3600, metadata={"description": "Max waiting time to complete a merge graph action."}
    )
    abort_after_seconds: int = field(
        default=4 * 3600,
        metadata={"description": "If a graph update takes longer than this duration, the update is aborted."},
    )
    keep_history: bool = field(
        default=True,
        metadata={"description": "If true, changes of the graph are stored and are available via history."},
    )
    parallel_imports: int = field(
        default=5,
        metadata={"description": "Number of parallel graph merge requests handled in parallel."},
    )

    def merge_max_wait_time(self) -> timedelta:
        return timedelta(seconds=self.merge_max_wait_time_seconds)

    def abort_after(self) -> timedelta:
        return timedelta(seconds=self.abort_after_seconds)


# Define rules to validate this config
schema_registry.add(
    schema_name(GraphUpdateConfig),
    dict(
        merge_max_wait_time_seconds={"type": "integer", "min": 60},
        abort_after_seconds={"type": "integer", "min": 60},
    ),
)


@define()
class RuntimeConfig(ConfigObject):
    kind: ClassVar[str] = f"{ResotoCoreRoot}_runtime_config"
    debug: bool = field(default=False, metadata={"description": "Enable debug logging and exception tracing."})
    log_level: str = field(default="info", metadata={"description": "Log level (default: info)"})
    plantuml_server: str = field(
        default="https://plantuml.resoto.org",
        metadata={"description": "PlantUML server URI for UML image rendering."},
    )
    start_collect_on_subscriber_connect: bool = field(
        default=False,
        metadata={"description": "Start the collect workflow, when the first handling actor connects to the system."},
    )
    usage_metrics: bool = field(
        default=True,
        metadata={
            "description": "Usage metrics provide information like errors and bugs, "
            "which we rely on to improve Resoto with every release.\n"
            "All metrics are anonymous. "
            "See https://resoto.com/docs/edge/reference/telemetry for more information.\n"
            "Please help us by leaving this setting turned on."
        },
    )


# Define rules to validate this config
schema_registry.add(
    schema_name(RuntimeConfig),
    dict(log_level={"type": "string", "allowed": ["critical", "fatal", "error", "warn", "warning", "info", "debug"]}),
)


@define()
class WorkflowConfig(ConfigObject):
    kind: ClassVar[str] = f"{ResotoCoreRoot}_workflow_config"
    schedule: str = field(metadata={"description": "Cron expression as schedule for the workflow to run."})


schema_registry.add(
    schema_name(WorkflowConfig),
    dict(schedule={"type": "string", "is_cron": True}),
)


@define()
class RunConfig(ConfigObject):
    temp_dir: Path = Path("/tmp")  # set to random temp directory during start of process
    verify: Union[bool, str, None] = None


@define()
class CoreConfig(ConfigObject):
    api: ApiConfig
    cli: CLIConfig
    graph_update: GraphUpdateConfig
    runtime: RuntimeConfig
    db: DatabaseConfig
    workflows: Dict[str, WorkflowConfig]
    custom_commands: CustomCommandsConfig
    snapshots: SnapshotsScheduleConfig
    args: Namespace
    run: RunConfig

    @property
    def editable(self) -> "EditableConfig":
        return EditableConfig(self.api, self.cli, self.graph_update, self.runtime, self.workflows)

    def json(self) -> Json:
        return {ResotoCoreRoot: to_js(self.editable, strip_attr="kind")}

    def validate(self) -> Optional[Json]:
        return self.editable.validate()


@define()
class EditableConfig(ConfigObject):
    kind: ClassVar[str] = ResotoCoreRoot
    api: ApiConfig = field(
        factory=ApiConfig,
        metadata={"description": "API related properties."},
    )
    cli: CLIConfig = field(
        factory=CLIConfig,
        metadata={"description": "CLI related properties."},
    )
    graph_update: GraphUpdateConfig = field(
        factory=GraphUpdateConfig,
        metadata={"description": "Properties for updating the graph."},
    )
    runtime: RuntimeConfig = field(
        factory=RuntimeConfig,
        metadata={"description": "Runtime related properties."},
    )
    workflows: Dict[str, WorkflowConfig] = field(
        factory=lambda: {"collect_and_cleanup": WorkflowConfig(schedule="0 * * * *")},
        metadata={"description": "Workflow related properties."},
    )


def config_model() -> List[Json]:
    config_classes = {EditableConfig, CustomCommandsConfig}
    return dataclasses_to_resotocore_model(config_classes, use_optional_as_required=True)


# Define rules to validate this config
# Note: since validation rules do not cover all attributes, we allow unknown properties explicitly.
schema_registry.add(
    schema_name(EditableConfig),
    dict(
        api={"schema": schema_name(ApiConfig), "allow_unknown": True},
        cli={"schema": schema_name(CLIConfig), "allow_unknown": True},
        graph_update={"schema": schema_name(GraphUpdateConfig), "allow_unknown": True},
        runtime={"schema": schema_name(RuntimeConfig), "allow_unknown": True},
        workflows={
            "type": "dict",
            "keysrules": {"type": "string"},
            "valuesrules": {"schema": schema_name(WorkflowConfig)},
        },
    ),
)


def parse_config(
    args: Namespace,
    core_config: Json,
    get_core_overrides: Callable[[], Optional[Json]],
    command_templates: Optional[Json] = None,
    snapshot_schedule: Optional[Json] = None,
) -> CoreConfig:
    db = DatabaseConfig(
        server=args.graphdb_server,
        database=args.graphdb_database,
        username=args.graphdb_username,
        password=args.graphdb_password,
        root_password=args.graphdb_root_password,
        bootstrap_do_not_secure=args.graphdb_bootstrap_do_not_secure,
        no_ssl_verify=args.graphdb_no_ssl_verify,
        request_timeout=args.graphdb_request_timeout,
    )
    # take command line options and translate it to the config model
    set_from_cmd_line = {
        "api.ui_path": args.ui_path,
        "runtime.debug": args.debug,
        "runtime.usage_metrics": not args.analytics_opt_out if args.analytics_opt_out is not None else None,
    }

    # take config overrides and adjust the configuration
    for key, value in args.config_override:
        set_from_cmd_line[ResotoCoreRootRE.sub("", key, 1)] = value

    # set the relevant value in the json config model
    migrated = migrate_core_config(core_config)
    adjusted = migrated.get(ResotoCoreRoot) or {}
    for path, value in set_from_cmd_line.items():
        if value is not None:
            adjusted = set_value_in_path(value, path, adjusted)

    # here we only care about the resotocore overrides
    core_config_overrides = (get_core_overrides() or {}).get(ResotoCoreRoot)
    # merge the file overrides into the adjusted config
    if core_config_overrides:
        adjusted = merge_json_elements(adjusted, core_config_overrides)

    # replacing the env vars and removing them in case they are not resolved
    adjusted = replace_env_vars(adjusted, os.environ, keep_unresolved=False)

    # coerce the resulting json to the config model
    try:
        model = Model.from_kinds(from_js(config_model(), List[Kind]))
        root = model.get(ResotoCoreRoot)
        if isinstance(root, ComplexKind):
            adjusted = root.coerce(adjusted)
    except Exception as e:
        log.warning(f"Can not adjust configuration: {e}", exc_info=e)

    try:
        # replace all env vars
        ed = from_js(adjusted, EditableConfig)
    except Exception as e:
        # only here as last resort - should never be required
        log.error("Final configuration can not be parsed! Fall back to default configuration.", exc_info=e)
        ed = EditableConfig()

    commands_config = CustomCommandsConfig()
    if command_templates:
        try:
            migrated_commands = migrate_command_config(command_templates)
            cmd_cfg_to_parse = migrated_commands or command_templates
            commands_config = from_js(cmd_cfg_to_parse.get(ResotoCoreCommandsRoot), CustomCommandsConfig)
        except Exception as e:
            log.error(f"Can not parse command templates. Fall back to defaults. Reason: {e}", exc_info=e)

    snapshots_config = SnapshotsScheduleConfig()
    if snapshot_schedule:
        try:
            snapshots_config = from_js(snapshot_schedule.get(ResotoCoreSnapshotsRoot), SnapshotsScheduleConfig)
        except Exception as e:
            log.error(f"Can not parse snapshot schedule. Fall back to defaults. Reason: {e}", exc_info=e)

    return CoreConfig(
        api=ed.api,
        args=args,
        cli=ed.cli,
        custom_commands=commands_config,
        snapshots=snapshots_config,
        db=db,
        graph_update=ed.graph_update,
        runtime=ed.runtime,
        workflows=ed.workflows,
        run=RunConfig(),  # overridden for each run
    )


def migrate_core_config(config: Json) -> Json:
    """
    :param config: The core configuration
    :return: the migrated json.
    """
    cfg = config.get(ResotoCoreRoot) or {}
    adapted = deepcopy(cfg)

    # 2.2 -> 2.3: rename and toggle `analytics_opt_out` -> `usage_metrics`
    opt_out = value_in_path(cfg, "runtime.analytics_opt_out")
    usage = value_in_path(cfg, "runtime.usage_metrics")
    if opt_out is not None and usage is None:
        set_value_in_path(not opt_out, "runtime.usage_metrics", adapted)
    del_value_in_path(adapted, "runtime.analytics_opt_out")

    # 3.0 -> 3.1: delete `api.ui_path`
    del_value_in_path(adapted, "api.ui_path")

    # 3.5 -> 3.6: web_port -> https_port
    if web_port := value_in_path(cfg, "api.web_port"):
        set_value_in_path(web_port, "api.https_port", adapted)
        del_value_in_path(adapted, "api.web_port")

    if value_in_path(cfg, "runtime.plantuml_server") == "http://plantuml.resoto.org:8080":
        set_value_in_path("https://plantuml.resoto.org", "runtime.plantuml_server", adapted)

    return {ResotoCoreRoot: adapted}


def migrate_command_config(cmd_config: Json) -> Optional[Json]:
    config = from_js(cmd_config.get(ResotoCoreCommandsRoot), CustomCommandsConfig)
    existing_commands = {tpl.name: tpl for tpl in config.commands}
    adjusted = False
    for command in alias_templates():
        if command.name not in existing_commands:
            config.commands.append(command)
            adjusted = True
    return config.json() if adjusted else None


def config_from_db(
    args: Namespace,
    db: StandardDatabase,
    get_core_overrides: Callable[[], Optional[Json]],
    collection_name: str = "configs",
) -> CoreConfig:
    if configs := db.collection(collection_name) if db.has_collection(collection_name) else None:
        if config_entity := cast(Optional[Json], configs.get(ResotoCoreConfigId)):
            if config := config_entity.get("config"):
                command_config_entity = cast(Optional[Json], configs.get(ResotoCoreCommandsConfigId))
                command_config = command_config_entity.get("config") if command_config_entity else None

                snapshots_config_entity = cast(Optional[Json], configs.get(ResotoCoreSnapshotsConfigId))
                snapshots_config = snapshots_config_entity.get("config") if snapshots_config_entity else None

                return parse_config(args, config, get_core_overrides, command_config, snapshots_config)
    return parse_config(args, {}, get_core_overrides)



from datetime import timedelta

from resotocore import version
from resotocore.analytics import AnalyticsEventSender, CoreEvent
from resotocore.message_bus import MessageBus
from resotocore.model.model_handler import ModelHandler
from resotocore.task.subscribers import SubscriptionHandler
from resotocore.util import Periodic
from resotocore.worker_task_queue import WorkerTaskQueue
from resotocore.ids import GraphName


def emit_recurrent_events(
    event_sender: AnalyticsEventSender,
    model_handler: ModelHandler,
    subscription_handler: SubscriptionHandler,
    worker_task_queue: WorkerTaskQueue,
    message_bus: MessageBus,
    frequency: timedelta,
    first_run: timedelta,
) -> Periodic:
    async def emit_events() -> None:
        # information about the model
        default_graph = GraphName("resoto")
        model = await model_handler.load_model(default_graph)
        await event_sender.core_event(
            CoreEvent.ModelInfo,
            dict(version=version()),
            model_count=len(model.kinds),
        )
        # information about all subscribers/actors
        subscribers = await subscription_handler.all_subscribers()
        await event_sender.core_event(
            CoreEvent.SubscriberInfo,
            subscriber_count=sum(1 for _ in subscribers),
            # do not count wildcard listeners
            active=sum(1 for channels in message_bus.active_listener.values() if channels != ["*"]),
        )
        # information about all workers
        await event_sender.core_event(
            CoreEvent.WorkerQueueInfo,
            worker_count=len(worker_task_queue.work_count),
            worker_tasks_count=len(worker_task_queue.worker_by_task_name),
            outstanding_tasks=len(worker_task_queue.outstanding_tasks),
            unassigned_tasks=len(worker_task_queue.unassigned_tasks),
        )

    return Periodic("emit_recurrent_events", emit_events, frequency, first_run)



from typing import Tuple, List

from resotocore.system_start import parse_args
from resotocore.types import JsonElement


def test_parse_override() -> None:
    def parse(args: str) -> List[Tuple[str, JsonElement]]:
        return parse_args(args.split()).config_override  # type: ignore

    assert parse(f"--override a=foo") == [("a", "foo")]
    assert parse(f"--override a=foo,bla") == [("a", ["foo", "bla"])]
    assert parse(f"--override a=foo,bla b=a,b,c") == [("a", ["foo", "bla"]), ("b", ["a", "b", "c"])]
    assert parse(f'--override a="value,with,comma,in,quotes"') == [("a", "value,with,comma,in,quotes")]
    assert parse(f'--override a=some,value,"with,comma"') == [("a", ["some", "value", "with,comma"])]



from typing import AsyncIterator, Optional, List, Dict, Union, Tuple, Callable
import asyncio
from asyncio import Lock
from pathlib import Path
import aiohttp
import aiofiles
from collections import defaultdict
from abc import ABC
from attrs import frozen
from jsons import loads as json_loads

from resotocore.infra_apps.manifest import AppManifest
from resotocore.db.packagedb import PackageEntityDb, InfraAppPackage
from resotocore.config import ConfigHandler
from resotocore.config import ConfigEntity
from resotocore.ids import InfraAppName, ConfigId
from resotocore.model.model import Kind, ComplexKind
from resotocore.types import Json
from resotocore.model.typed_model import from_js
from resotocore.cli.model import InfraAppAlias, InfraAppAliasParameter
from logging import getLogger
from resotocore.service import Service

logger = getLogger(__name__)


def config_id(name: InfraAppName) -> ConfigId:
    return ConfigId(f"resoto.apps.{name}")


class Failure(ABC):
    def __nonzero__(self) -> bool:
        return False

    def __bool__(self) -> bool:
        return False


@frozen
class ManifestNotFound(Failure):
    name: InfraAppName

    def __str__(self) -> str:
        return f"Manifest for app {self.name} was not found."


@frozen
class ManifestDownloadFailed(Failure):
    url: str
    reason: str

    def __str__(self) -> str:
        return f"Failed to download app manifest from {self.url}: {self.reason}"


@frozen
class ManifestInstallFailed(Failure):
    name: InfraAppName
    reason: str

    def __str__(self) -> str:
        return f"Failed to install app {self.name} from manifest. Reason: {self.reason}"


class PackageManager(Service):
    def __init__(
        self,
        entity_db: PackageEntityDb,
        config_handler: ConfigHandler,
        add_command_alias: Callable[[InfraAppAlias], None],
        remove_command_alias: Callable[[str], None],
        repos_cache_directory: Path = Path.home() / ".cache" / "resoto-infra-apps",
    ) -> None:
        """
        Package manager for infra apps.

        Args:
            entity_db: EntityDb for storing app manifests
            config_handler: ConfigHandler for storing app configs
            repos_cache_directory: Directory where the infra apps repos are cloned
        """

        self.entity_db = entity_db
        self.config_handler = config_handler
        self.update_lock: Optional[Lock] = None
        self.repos_cache_directory: Path = repos_cache_directory
        self.cleanup_task: Optional[asyncio.Task[None]] = None
        self.cdn_url = "https://cdn.some.engineering/resoto/apps/index.json"
        self.add_command_alias = add_command_alias
        self.remove_command_alias = remove_command_alias

    async def start(self) -> None:
        self.update_lock = asyncio.Lock()
        self.repos_cache_directory.mkdir(parents=True, exist_ok=True)

        # set up custom commands
        manifests = [await self.get_manifest(name) async for name in self.list()]
        self._setup_command_aliases([m for m in manifests if m is not None])

    def _setup_command_aliases(self, manifests: List[AppManifest]) -> None:
        for manifest in manifests:

            def to_param(arg: Tuple[str, Json]) -> InfraAppAliasParameter:
                name, arg_schema = arg
                return InfraAppAliasParameter(
                    name=name,
                    help=arg_schema.get("help", ""),
                    default=arg_schema.get("default", None),
                )

            parameters = [to_param(arg) for arg in (manifest.args_schema or {}).items()]
            alias_template = InfraAppAlias(
                name=manifest.name,
                description=manifest.description,
                readme=manifest.readme,
                parameters=parameters,
            )
            self.add_command_alias(alias_template)

    async def stop(self) -> None:
        if self.cleanup_task:
            self.cleanup_task.cancel()
            await self.cleanup_task

    async def search(self, query: Optional[str], url: Optional[str]) -> AsyncIterator[AppManifest]:
        maybe_available_manifests = await self._get_manifests(url)
        if isinstance(maybe_available_manifests, Failure):
            raise RuntimeError(str(maybe_available_manifests))
        available_manifests = maybe_available_manifests.values()
        if query is None:
            for manifest in available_manifests:
                yield manifest
        else:
            query = query.lower()
            for manifest in available_manifests:
                if query in manifest.name.lower():
                    yield manifest
                    continue
                if query in manifest.description.lower():
                    yield manifest
                    continue
                if query in manifest.readme.lower():
                    yield manifest
                    continue
                if query in manifest.categories:
                    yield manifest
                    continue

    def list(self) -> AsyncIterator[InfraAppName]:
        return self.entity_db.keys()

    async def get_manifest(self, name: InfraAppName) -> Optional[AppManifest]:
        if package := await self.entity_db.get(name):
            return package.manifest
        return None

    async def info(self, name: InfraAppName, url: Optional[str]) -> Json:
        async def get_manifest() -> AppManifest:
            if package := await self.entity_db.get(name):
                return package.manifest
            else:
                result = await self._fetch_manifest(name, url or self.cdn_url)
                if isinstance(result, Failure):
                    raise ValueError(f"Can't find the package {name}, reason: {result}")
                return result

        manifest = await get_manifest()
        json = {
            "name": manifest.name,
            "description": manifest.description,
            "url": manifest.url,
            "version": manifest.version,
        }
        return json

    async def update(self, name: InfraAppName) -> AppManifest:
        assert self.update_lock, "PackageManager not started"

        async with self.update_lock:
            if package := await self.entity_db.get(name):
                new_manifest = await self._fetch_manifest(name, package.source_url)
                if isinstance(new_manifest, Failure):
                    logger.warning(f"Failed to fetch manifest for app {name}, skipping update")
                    raise RuntimeError(str(new_manifest))

                await self._delete(package.manifest.name)
                result = await self._install_from_manifest(new_manifest, package.source_url)
                if isinstance(result, Failure):
                    logger.warning(f"Failed to install app {name} from manifest, skipping update")
                    raise RuntimeError(str(result))
                return result
            else:
                raise RuntimeError(f"App {name} is not installed, cannot update")

    async def update_from_manifest(self, manifest: AppManifest) -> AppManifest:
        assert self.update_lock, "PackageManager not started"

        async with self.update_lock:
            if package := await self.entity_db.get(manifest.name):
                await self._delete(package.manifest.name)
                result = await self._install_from_manifest(manifest, package.source_url)
                if isinstance(result, Failure):
                    logger.warning(f"Failed to install app {manifest.name} from manifest, skipping update")
                    raise RuntimeError(str(result))
                return result
            else:
                raise RuntimeError(f"App {manifest.name} is not installed, cannot update")

    async def update_all(self) -> AsyncIterator[Tuple[InfraAppName, Union[AppManifest, Failure]]]:
        assert self.update_lock, "PackageManager not started"

        async with self.update_lock:
            packages: Dict[str, List[AppManifest]] = defaultdict(list)
            async for package in self.entity_db.all():
                packages[package.source_url].append(package.manifest)

            for url, manifests in packages.items():
                manifests_cache = await self._get_manifests(url)
                if isinstance(manifests_cache, Failure):
                    logger.warning(f"Failed to download manifests for url {url}, skipping update")
                    continue
                for manifest in manifests:
                    new_manifest = manifests_cache.get(manifest.name, ManifestNotFound(manifest.name))
                    if isinstance(new_manifest, Failure):
                        logger.warning(f"Failed to download manifest for app {manifest.name}, skipping update")
                        yield (manifest.name, new_manifest)
                        continue
                    await self._delete(manifest.name)
                    await self._install_from_manifest(new_manifest, url)
                    yield (manifest.name, new_manifest)

    async def delete(self, name: InfraAppName) -> None:
        assert self.update_lock, "PackageManager not started"
        async with self.update_lock:
            await self._delete(name)

    async def _delete(self, name: InfraAppName) -> None:
        if await self.entity_db.get(name) is None:
            return
        await self.entity_db.delete(name)
        # clean up the aliases
        self.remove_command_alias(name)

    # user-facing method, errors are thrown as exceptions
    async def install(self, name: InfraAppName, url: Optional[str]) -> AppManifest:
        assert self.update_lock, "PackageManager not started"

        async with self.update_lock:
            if installed := await self.entity_db.get(name):
                return installed.manifest

            manifest = await self._fetch_manifest(name, url or self.cdn_url)
            if isinstance(manifest, Failure):
                raise RuntimeError(str(manifest))
            result = await self._install_from_manifest(manifest, url or self.cdn_url)
            if isinstance(result, Failure):
                raise RuntimeError(str(result))
            return result

    async def _install_from_manifest(
        self, manifest: AppManifest, url: str
    ) -> Union[ManifestInstallFailed, AppManifest]:
        conf_id = config_id(manifest.name)
        if schema := manifest.config_schema:
            kinds: List[Kind] = [from_js(kind, ComplexKind) for kind in schema]
            await self.config_handler.update_configs_model(kinds)
        config_entity = ConfigEntity(conf_id, manifest.default_config or {}, None)
        try:
            if await self.config_handler.get_config(conf_id) is None:
                await self.config_handler.put_config(config_entity, validate=manifest.config_schema is not None)
            stored = await self.entity_db.update(InfraAppPackage(manifest, url))

            # add the custom command alias
            self._setup_command_aliases([manifest])

            return stored.manifest
        except Exception as e:
            logger.error(f"Failed to install {manifest.name} from {url}", exc_info=e)
            return ManifestInstallFailed(manifest.name, str(e))

    async def _fetch_manifest(self, app_name: InfraAppName, url: str) -> Union[Failure, AppManifest]:
        url = url or self.cdn_url
        manifests = await self._get_manifests(url)
        if isinstance(manifests, Failure):
            return manifests
        return manifests.get(app_name, ManifestNotFound(app_name))

    async def _get_manifests(self, url: Optional[str]) -> Union[Failure, Dict[InfraAppName, AppManifest]]:
        url = url or self.cdn_url
        if url.startswith("file://"):
            path = Path(url[7:])
            return await self._get_manifests_from_file(path)
        else:
            return await self._get_manifests_from_cdn(url)

    async def _get_manifests_from_cdn(self, url: str) -> Union[Failure, Dict[InfraAppName, AppManifest]]:
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as response:
                    if response.status != 200:
                        raise RuntimeError(f"Failed to fetch {self.cdn_url}")
                    manifests_bytes = await response.read()
                    json = json_loads(manifests_bytes.decode())
                    assert isinstance(json, list)
                    manifests: Dict[InfraAppName, AppManifest] = {}
                    for manifest_json in json:
                        manifest = AppManifest.from_json(manifest_json)
                        manifests[manifest.name] = manifest
                    return manifests
        except Exception as e:
            logger.error(f"Failed to fetch manifests from {self.cdn_url}", exc_info=e)
            return ManifestDownloadFailed(url, str(e))

    async def _get_manifests_from_file(self, path: Path) -> Union[Failure, Dict[InfraAppName, AppManifest]]:
        try:
            async with aiofiles.open(path) as f:
                manifests_bytes = await f.read()
                json = json_loads(manifests_bytes)
                assert isinstance(json, list)
                manifests: Dict[InfraAppName, AppManifest] = {}
                for manifest_json in json:
                    manifest = AppManifest.from_json(manifest_json)
                    manifests[manifest.name] = manifest
                return manifests
        except Exception as e:
            logger.error(f"Failed to fetch manifests from {path}", exc_info=e)
            return ManifestDownloadFailed(str(path.absolute()), str(e))



from typing import List, Optional, Dict
from attrs import frozen
from resotocore.types import Json
from resotocore.model.typed_model import from_js
from resotocore.ids import InfraAppName
from yaml import safe_load
from jsons import loads as jsons_loads


ArgName = str


@frozen
class AppManifest:
    """
    The manifest of an infrastructure app. The manifest is a YAML file that describes the infrastsucture app.

    @param name: The name of the app. Acts as the unique identifier of the app. Allowed characters are [a-zA-Z0-9_].
    @param description: A short human-readable description of the app.
    @param version: The version of the app. Must be a valid semantic version.
    @param readme: A long human-readable description of the app in markdown format.
    @param language: The programming language the app is written. Currently only "jinja2" is supported.
    @param url: The URL to the app's source code location.
    @param icon: base64 encoded icon of the app.
    @param categories: A list of categories the app belongs to.
    @param config_schema: A JSON schema that describes the configuration of the app.
    See https://json-schema.org/ for more information.
    @param default_config: The default configuration of the app.
    @param args_schema: A list of command line arguments that can be passed to the app.
    Must be compatible with the Python argparse library.
    @param source: The source code of the app.

    """

    name: InfraAppName
    description: str
    version: str
    readme: str
    language: str
    url: str
    icon: str
    categories: List[str]
    config_schema: Optional[List[Json]]
    default_config: Optional[Json]
    args_schema: Optional[Dict[ArgName, Json]]
    source: str

    # Object creation methods. Use these instead of the __init__ method.

    @staticmethod
    def from_json(json: Json) -> "AppManifest":
        return from_js(json, AppManifest)

    @staticmethod
    def from_json_str(json_str: str) -> "AppManifest":
        json = jsons_loads(json_str)
        manifest = AppManifest.from_json(json)
        return manifest

    @staticmethod
    def from_bytes(b: bytes) -> "AppManifest":
        return AppManifest.from_json_str(b.decode("utf-8"))

    @staticmethod
    def from_yaml_str(yaml_str: str) -> "AppManifest":
        return AppManifest.from_json(safe_load(yaml_str))
